from unittest import mock

import pytest
from django.urls import reverse_lazy


@pytest.mark.parametrize(
    "header, incoming_ip, expected",
    (
        ("HTTP_X_FORWARDED_FOR", "74.125.224.72", 200),
        ("X_FORWARDED_FOR", "74.125.224.72", 200),
        ("HTTP_CLIENT_IP", "74.125.224.72", 200),
        ("HTTP_X_REAL_IP", "74.125.224.72", 200),
        ("HTTP_X_FORWARDED", "74.125.224.72", 200),
        ("HTTP_X_CLUSTER_CLIENT_IP", "74.125.224.72", 200),
        ("HTTP_FORWARDED_FOR", "74.125.224.72", 200),
        ("HTTP_FORWARDED", "74.125.224.72", 200),
        ("HTTP_VIA", "74.125.224.72", 200),
        ("REMOTE_ADDR", "74.125.224.72", 200),
    ),
    ids=[
        "HTTP_X_FORWARDED_FOR allow",
        "X_FORWARDED_FOR allow",
        "HTTP_CLIENT_IP allow",
        "HTTP_X_REAL_IP allow",
        "HTTP_X_FORWARDED allow",
        "HTTP_X_CLUSTER_CLIENT_IP allow",
        "HTTP_FORWARDED_FOR allow",
        "HTTP_FORWARDED allow",
        "HTTP_VIA allow",
        "REMOTE_ADDR allow",
    ],
)
def test_admin_no_restriction(header, incoming_ip, expected, settings, client):
    settings.RESTRICT_ADMIN = False
    admin_url = reverse_lazy("admin:login")
    response = client.get(admin_url, **{header: incoming_ip})
    assert response.status_code == expected


@pytest.mark.parametrize(
    "header, incoming_ip, expected",
    (
        ("HTTP_X_FORWARDED_FOR", "74.125.224.72", 200),
        ("X_FORWARDED_FOR", "74.125.224.72", 200),
        ("HTTP_CLIENT_IP", "74.125.224.72", 200),
        ("HTTP_X_REAL_IP", "74.125.224.72", 200),
        ("HTTP_X_FORWARDED", "74.125.224.72", 200),
        ("HTTP_X_CLUSTER_CLIENT_IP", "74.125.224.72", 200),
        ("HTTP_FORWARDED_FOR", "74.125.224.72", 200),
        ("HTTP_FORWARDED", "74.125.224.72", 200),
        ("REMOTE_ADDR", "74.125.224.72", 200),
        ("HTTP_X_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("X_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_CLIENT_IP", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_X_REAL_IP", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_X_FORWARDED", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_X_CLUSTER_CLIENT_IP", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_FORWARDED", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("REMOTE_ADDR", "0:0:0:0:0:ffff:4a7d:e048", 200),
    ),
    ids=[
        "HTTP_X_FORWARDED_FOR allow",
        "X_FORWARDED_FOR allow",
        "HTTP_CLIENT_IP allow",
        "HTTP_X_REAL_IP allow",
        "HTTP_X_FORWARDED allow",
        "HTTP_X_CLUSTER_CLIENT_IP allow",
        "HTTP_FORWARDED_FOR allow",
        "HTTP_FORWARDED allow",
        "REMOTE_ADDR allow",
        "HTTP_X_FORWARDED_FOR ipv6 allow",
        "X_FORWARDED_FOR ipv6 allow",
        "HTTP_CLIENT_IP ipv6 allow",
        "HTTP_X_REAL_IP ipv6 allow",
        "HTTP_X_FORWARDED ipv6 allow",
        "HTTP_X_CLUSTER_CLIENT_IP ipv6 allow",
        "HTTP_FORWARDED_FOR ipv6 allow",
        "HTTP_FORWARDED ipv6 allow",
        "REMOTE_ADDR ipv6 allow",
    ],
)
def test_admin_restricted_allowed_ips(header, incoming_ip, expected, settings, client):
    settings.RESTRICT_ADMIN = True
    settings.ALLOWED_ADMIN_IPS = ["74.125.224.72", "0:0:0:0:0:ffff:4a7d:e048"]
    admin_url = reverse_lazy("admin:login")
    response = client.get(admin_url, **{header: incoming_ip})
    assert response.status_code == expected


@pytest.mark.parametrize(
    "header, incoming_ip, expected",
    (
        ("HTTP_X_FORWARDED_FOR", "74.125.224.72", 404),
        ("X_FORWARDED_FOR", "74.125.224.72", 404),
        ("HTTP_CLIENT_IP", "74.125.224.72", 404),
        ("HTTP_X_REAL_IP", "74.125.224.72", 404),
        ("HTTP_X_FORWARDED", "74.125.224.72", 404),
        ("HTTP_X_CLUSTER_CLIENT_IP", "74.125.224.72", 404),
        ("HTTP_FORWARDED_FOR", "74.125.224.72", 404),
        ("HTTP_FORWARDED", "74.125.224.72", 404),
        ("REMOTE_ADDR", "74.125.224.72", 404),
        ("HTTP_X_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("X_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_CLIENT_IP", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_X_REAL_IP", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_X_FORWARDED", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_X_CLUSTER_CLIENT_IP", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_FORWARDED", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("REMOTE_ADDR", "0:0:0:0:0:ffff:4a7d:e048", 404),
    ),
    ids=[
        "HTTP_X_FORWARDED_FOR block",
        "X_FORWARDED_FOR block",
        "HTTP_CLIENT_IP block",
        "HTTP_X_REAL_IP block",
        "HTTP_X_FORWARDED block",
        "HTTP_X_CLUSTER_CLIENT_IP block",
        "HTTP_FORWARDED_FOR block",
        "HTTP_FORWARDED block",
        "REMOTE_ADDR block",
        "HTTP_X_FORWARDED_FOR ipv6 block",
        "X_FORWARDED_FOR ipv6 block",
        "HTTP_CLIENT_IP ipv6 block",
        "HTTP_X_REAL_IP ipv6 block",
        "HTTP_X_FORWARDED ipv6 block",
        "HTTP_X_CLUSTER_CLIENT_IP ipv6 block",
        "HTTP_FORWARDED_FOR ipv6 block",
        "HTTP_FORWARDED ipv6 block",
        "REMOTE_ADDR ipv6 block",
    ],
)
def test_admin_restricted_blocked_ips(header, incoming_ip, expected, settings, client):
    settings.RESTRICT_ADMIN = True
    settings.ALLOWED_ADMIN_IPS = ["8.8.8.9", "0:0:0:0:0:ffff:4a7d:e04b"]
    admin_url = reverse_lazy("admin:login")
    response = client.get(admin_url, **{header: incoming_ip})
    assert response.status_code == expected


@pytest.mark.parametrize(
    "header, incoming_ip, expected",
    (
        ("HTTP_X_FORWARDED_FOR", "74.125.224.72", 200),
        ("X_FORWARDED_FOR", "74.125.224.72", 200),
        ("HTTP_CLIENT_IP", "74.125.224.72", 200),
        ("HTTP_X_REAL_IP", "74.125.224.72", 200),
        ("HTTP_X_FORWARDED", "74.125.224.72", 200),
        ("HTTP_X_CLUSTER_CLIENT_IP", "74.125.224.72", 200),
        ("HTTP_FORWARDED_FOR", "74.125.224.72", 200),
        ("HTTP_FORWARDED", "74.125.224.72", 200),
        ("REMOTE_ADDR", "74.125.224.72", 200),
        ("HTTP_X_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("X_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_CLIENT_IP", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_X_REAL_IP", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_X_FORWARDED", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_X_CLUSTER_CLIENT_IP", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("HTTP_FORWARDED", "0:0:0:0:0:ffff:4a7d:e048", 200),
        ("REMOTE_ADDR", "0:0:0:0:0:ffff:4a7d:e048", 200),
    ),
    ids=[
        "HTTP_X_FORWARDED_FOR allow",
        "X_FORWARDED_FOR allow",
        "HTTP_CLIENT_IP allow",
        "HTTP_X_REAL_IP allow",
        "HTTP_X_FORWARDED allow",
        "HTTP_X_CLUSTER_CLIENT_IP allow",
        "HTTP_FORWARDED_FOR allow",
        "HTTP_FORWARDED allow",
        "REMOTE_ADDR allow",
        "HTTP_X_FORWARDED_FOR ipv6 allow",
        "X_FORWARDED_FOR ipv6 allow",
        "HTTP_CLIENT_IP ipv6 allow",
        "HTTP_X_REAL_IP ipv6 allow",
        "HTTP_X_FORWARDED ipv6 allow",
        "HTTP_X_CLUSTER_CLIENT_IP ipv6 allow",
        "HTTP_FORWARDED_FOR ipv6 allow",
        "HTTP_FORWARDED ipv6 allow",
        "REMOTE_ADDR ipv6 allow",
    ],
)
def test_admin_restricted_allowed_ip_ranges(header, incoming_ip, expected, settings, client):
    settings.RESTRICT_ADMIN = True
    settings.ALLOWED_ADMIN_IP_RANGES = ["74.125.224.72/32", "0:0:0:0:0:ffff:4a7d:e048/128"]
    admin_url = reverse_lazy("admin:login")
    response = client.get(admin_url, **{header: incoming_ip})
    assert response.status_code == expected


@pytest.mark.parametrize(
    "header, incoming_ip, expected",
    (
        ("HTTP_X_FORWARDED_FOR", "74.125.224.72", 404),
        ("X_FORWARDED_FOR", "74.125.224.72", 404),
        ("HTTP_CLIENT_IP", "74.125.224.72", 404),
        ("HTTP_X_REAL_IP", "74.125.224.72", 404),
        ("HTTP_X_FORWARDED", "74.125.224.72", 404),
        ("HTTP_X_CLUSTER_CLIENT_IP", "74.125.224.72", 404),
        ("HTTP_FORWARDED_FOR", "74.125.224.72", 404),
        ("HTTP_FORWARDED", "74.125.224.72", 404),
        ("REMOTE_ADDR", "74.125.224.72", 404),
        ("HTTP_X_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("X_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_CLIENT_IP", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_X_REAL_IP", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_X_FORWARDED", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_X_CLUSTER_CLIENT_IP", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_FORWARDED_FOR", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("HTTP_FORWARDED", "0:0:0:0:0:ffff:4a7d:e048", 404),
        ("REMOTE_ADDR", "0:0:0:0:0:ffff:4a7d:e048", 404),
    ),
    ids=[
        "HTTP_X_FORWARDED_FOR block",
        "X_FORWARDED_FOR block",
        "HTTP_CLIENT_IP block",
        "HTTP_X_REAL_IP block",
        "HTTP_X_FORWARDED block",
        "HTTP_X_CLUSTER_CLIENT_IP block",
        "HTTP_FORWARDED_FOR block",
        "HTTP_FORWARDED block",
        "REMOTE_ADDR block",
        "HTTP_X_FORWARDED_FOR ipv6 block",
        "X_FORWARDED_FOR ipv6 block",
        "HTTP_CLIENT_IP ipv6 block",
        "HTTP_X_REAL_IP ipv6 block",
        "HTTP_X_FORWARDED ipv6 block",
        "HTTP_X_CLUSTER_CLIENT_IP ipv6 block",
        "HTTP_FORWARDED_FOR ipv6 block",
        "HTTP_FORWARDED ipv6 block",
        "REMOTE_ADDR ipv6 block",
    ],
)
def test_admin_restricted_blocked_ip_ranges(header, incoming_ip, expected, settings, client):
    settings.RESTRICT_ADMIN = True
    settings.ALLOWED_ADMIN_IPS = ["192.168.0.0/24", "0:0:0:0:0:ffff:4a7d:e04b"]
    admin_url = reverse_lazy("admin:login")
    response = client.get(admin_url, **{header: incoming_ip})
    assert response.status_code == expected


@mock.patch("admin_ip_restrictor.middleware.get_client_ip")
def test_client_ip_not_found(mocked_get_client_ip, settings, client):
    settings.RESTRICT_ADMIN = True
    mocked_get_client_ip.return_value = None, None

    admin_url = reverse_lazy("admin:login")
    with pytest.raises(Exception) as e:
        client.get(admin_url)
    assert "IP not found" in str(e.value)


@mock.patch("admin_ip_restrictor.middleware.get_client_ip")
def test_ip_not_foud_unrestricted_app(mocked_get_client_ip, settings, client):
    settings.RESTRICT_ADMIN = True
    settings.RESTRICTED_APP_NAMES = ["foo"]
    mocked_get_client_ip.return_value = None, None

    admin_url = reverse_lazy("not-admin:home")

    response = client.get(admin_url)

    assert response.status_code == 200


@pytest.mark.parametrize(
    "header, incoming_ip",
    (
        ("HTTP_X_FORWARDED_FOR", "127.0.0.1"),
        ("HTTP_X_FORWARDED_FOR", "fc00:"),
    ),
    ids=["Private IPV4", "Private IPV6"],
)
def test_client_ip_private(header, incoming_ip, settings, client):
    settings.RESTRICT_ADMIN = True
    admin_url = reverse_lazy("admin:login")
    with pytest.raises(Exception) as e:
        client.get(admin_url, **{header: incoming_ip})
    assert "IP is private" in str(e.value)


@pytest.mark.parametrize(
    "header, incoming_ip",
    (
        ("HTTP_X_FORWARDED_FOR", "127.0.0.1"),
        ("HTTP_X_FORWARDED_FOR", "fc00:"),
    ),
    ids=["Private IPV4", "Private IPV6"],
)
def test_client_trust_ip_private(header, incoming_ip, settings, client):
    settings.RESTRICT_ADMIN = True
    settings.TRUST_PRIVATE_IP = True
    admin_url = reverse_lazy("admin:login")

    response = client.get(admin_url, **{header: incoming_ip})

    assert response.status_code == 200


@pytest.mark.parametrize(
    "envar, expected",
    (
        ("127.0.0.1", ["127.0.0.1"]),
        ("127.0.0.1,127.0.0.2", ["127.0.0.1", "127.0.0.2"]),
    ),
    ids=[
        "single entry string",
        "comma separated multiple entry string",
    ],
)
def test_list_envars_parsing(envar, expected):
    from admin_ip_restrictor.middleware import AdminIPRestrictorMiddleware

    assert AdminIPRestrictorMiddleware.parse_list_envars(envar) == expected


@pytest.mark.parametrize(
    "envar, expected",
    (("True", True), ("true", True), ("1", True), (1, True), ("bla", False), ("foo", False), ("0", False), (0, False)),
    ids=["True", "true", "1", "1 number", "false", "False", "0", "0 number"],
)
def test_bool_envars_parsing(envar, expected):
    from admin_ip_restrictor.middleware import AdminIPRestrictorMiddleware

    assert AdminIPRestrictorMiddleware.parse_bool_envars(envar) == expected



from django import VERSION as DJANGO_VERSION

SECRET_KEY = "fake-key"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.admin",
    "tests",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

ROOT_URLCONF = "tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]


if DJANGO_VERSION < (1, 10):
    MIDDLEWARE_CLASSES = [
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
        "admin_ip_restrictor.middleware.AdminIPRestrictorMiddleware",
    ]
else:
    MIDDLEWARE = [
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
        "admin_ip_restrictor.middleware.AdminIPRestrictorMiddleware",
    ]



import ipaddress

from django.conf import settings
from django.http import Http404
from django.utils.deprecation import MiddlewareMixin
from ipware import get_client_ip


class AdminIPRestrictorMiddleware(MiddlewareMixin):
    def __init__(self, get_response=None):
        super().__init__(get_response)
        restrict_admin = getattr(settings, "RESTRICT_ADMIN", False)
        trust_private_ip = getattr(settings, "TRUST_PRIVATE_IP", False)
        self.trust_private_ip = self.parse_bool_envars(trust_private_ip)
        self.restrict_admin = self.parse_bool_envars(restrict_admin)
        allowed_admin_ips = getattr(settings, "ALLOWED_ADMIN_IPS", [])
        self.allowed_admin_ips = self.parse_list_envars(allowed_admin_ips)
        allowed_admin_ip_ranges = getattr(settings, "ALLOWED_ADMIN_IP_RANGES", [])
        self.allowed_admin_ip_ranges = self.parse_list_envars(allowed_admin_ip_ranges)
        restricted_app_names = getattr(settings, "RESTRICTED_APP_NAMES", [])
        self.restricted_app_names = self.parse_list_envars(restricted_app_names)
        self.restricted_app_names.append("admin")

    @staticmethod
    def parse_bool_envars(value):
        return value in ("true", "True", "1", 1)

    @staticmethod
    def parse_list_envars(value):
        if isinstance(value, list):
            return value
        else:
            return value.split(",")

    def is_blocked(self, ip):
        """Determine if an IP address should be considered blocked."""
        blocked = True

        if self.trust_private_ip and ipaddress.ip_address(ip).is_private:
            blocked = False

        if ip in self.allowed_admin_ips:
            blocked = False

        for allowed_range in self.allowed_admin_ip_ranges:
            if ipaddress.ip_address(ip) in ipaddress.ip_network(allowed_range):
                blocked = False

        return blocked

    def get_ip(self, request):
        client_ip, is_routable = get_client_ip(request)
        assert client_ip, "IP not found"
        if not self.trust_private_ip:
            assert is_routable, "IP is private"
        return client_ip

    def process_view(self, request, view_func, view_args, view_kwargs):
        app_name = request.resolver_match.app_name
        is_restricted_app = app_name in self.restricted_app_names

        if self.restrict_admin and is_restricted_app:
            ip = self.get_ip(request)
            if self.is_blocked(ip):
                raise Http404()

        return None



from django.contrib import admin
from django.http.response import HttpResponse
from django.urls import include, path


def my_view(*args, **kwargs):
    return HttpResponse("Hello, world")


not_admin_urls = [
    path("", my_view, name="home"),
]

urlpatterns = [
    path("admin/", admin.site.urls),
    path("not-admin/", include((not_admin_urls, "not-admin"))),
]
